@@ -23,6 +23,9 @@
 #include "timing.h"
 #include "freelist.h"
 #include "mem.h"
+#include "tls.h"
+
+#define SUPPRESS_WARNING(a) (void)a
 
 #ifdef IS_PY3K
 PyDoc_STRVAR(_yappi__doc__, "Yet Another Python Profiler");
@@ -82,6 +85,11 @@ typedef struct {
     _pit_children_info *children;
 } _pit; // profile_item
 
+typedef struct {
+    int paused;
+    long long paused_at;
+} _glstate;
+
 typedef struct {
     _cstack *cs;
     _htab *rec_levels;
@@ -103,7 +111,11 @@ typedef struct {
     // how many times this thread is scheduled
     unsigned long sched_cnt;
 
+    long long last_seen;
+
     PyThreadState *ts_ptr;
+
+    _glstate gl_state;
 } _ctx; // context
 
 typedef struct {
@@ -111,25 +123,32 @@ typedef struct {
     PyObject *tag;
     PyObject *name;
     PyObject *modname;
-} _func_stat_filter;
+} _fast_func_stat_filter;
 
 typedef struct
 {
-    _func_stat_filter func_filter;
+    _fast_func_stat_filter func_filter;
     PyObject *enumfn;
 } _ctxenumarg;
 
 typedef struct
 {
     _ctxenumarg *enum_args;
+    uintptr_t tag;
     _ctx *ctx;
 } _ctxfuncenumarg;
 
 typedef struct {
     int builtins;
-    int multithreaded;
+    int multicontext;
 } _flag; // flags passed from yappi.start()
 
+typedef enum
+{
+    NATIVE_THREAD = 0x00,
+    GREENLET = 0x01,
+} _ctx_type_t;
+
 // globals
 static PyObject *YappiProfileError;
 static _htab *contexts;
@@ -145,6 +164,7 @@ static int paused;
 static time_t yappstarttime;
 static long long yappstarttick;
 static long long yappstoptick;
+static tls_key_t* tl_prev_ctx_key = NULL;
 static _ctx *prev_ctx = NULL;
 static _ctx *current_ctx = NULL;
 static _ctx *initial_ctx = NULL; // used for holding the context that called start()
@@ -153,6 +173,7 @@ static PyObject *tag_callback = NULL;
 static PyObject *context_name_callback = NULL;
 static PyObject *test_timings; // used for testing
 static const uintptr_t DEFAULT_TAG = 0;
+static _ctx_type_t ctx_type = NATIVE_THREAD;
 
 // defines
 #define UNINITIALIZED_STRING_VAL "N/A"
@@ -170,13 +191,15 @@ static const uintptr_t DEFAULT_TAG = 0;
 #endif
 
 #define PyLong_AsVoidPtr (uintptr_t)PyLong_AsVoidPtr
-
+ 
 // forwards
 static _ctx * _profile_thread(PyThreadState *ts);
+static void _pause_greenlet_ctx(_ctx *ctx);
+static void _resume_greenlet_ctx(_ctx *ctx);
 static int _pitenumdel(_hitem *item, void *arg);
 
 // funcs
-/*
+
 static void _DebugPrintObjects(unsigned int arg_count, ...)
 {
     unsigned int i;
@@ -188,7 +211,17 @@ static void _DebugPrintObjects(unsigned int arg_count, ...)
     }
     printf("\n");
     va_end(vargs);
-} */
+}
+
+int 
+IS_SUSPENDED(PyFrameObject *frame)
+{
+#if PY_MAJOR_VERSION >= 3 && PY_MINOR_VERSION >= 10
+    return (frame->f_state == FRAME_SUSPENDED);
+#else
+    return (frame->f_stacktop != NULL);
+#endif
+}
 
 int IS_ASYNC(PyFrameObject *frame)
 {
@@ -228,7 +261,7 @@ _log_err(unsigned int code)
 }
 
 static _pit *
-_create_pit()
+_create_pit(void)
 {
     _pit *pit;
 
@@ -273,6 +306,7 @@ _create_ctx(void)
     ctx->tid = 0;
     ctx->name = NULL;
     ctx->t0 = tickcount();
+    ctx->last_seen = ctx->t0;
     ctx->rec_levels = htcreate(HT_RLEVEL_SIZE);
     if (!ctx->rec_levels)
         return NULL;
@@ -396,7 +430,7 @@ _current_context_id(PyThreadState *ts)
         // thread. Therefore, we need to identify the "context" concept independent from ThreadState 
         // objects.
 
-        if (!flags.multithreaded) {
+        if (!flags.multicontext) {
             return 0;
         }
 
@@ -437,6 +471,8 @@ _thread2ctx(PyThreadState *ts)
         // callback functions in some circumtances, can be called before the context entry is not
         // created. (See issue 21). To prevent this problem we need to ensure the context entry for
         // the thread is always available here.
+        //
+        // This path is also excercised when new greenlets are encountered on an already profiled thread.
         return _profile_thread(ts);
     }
     return (_ctx *)it->val;
@@ -523,7 +559,7 @@ _ccode2pit(void *cco, uintptr_t current_tag)
 {
     PyCFunctionObject *cfn;
     _hitem *it;
-    PyObject *name;
+    PyObject *name, *mo, *obj_type, *method_descriptor;
     _htab *pits;
 
     pits = _get_pits_tbl(current_tag);
@@ -547,15 +583,25 @@ _ccode2pit(void *cco, uintptr_t current_tag)
         pit->builtin = 1;
         pit->modname = _pycfunction_module_name(cfn);
         pit->lineno = 0;
-        pit->fn_descriptor = (PyObject *)cfn;
-        Py_INCREF(cfn);
+        pit->fn_descriptor = NULL;
 
         // built-in method?
         if (cfn->m_self != NULL) {
             name = PyStr_FromString(cfn->m_ml->ml_name);
             if (name != NULL) {
-                PyObject *obj_type = PyObject_Type(cfn->m_self);
-                PyObject *mo = _PyType_Lookup((PyTypeObject *)obj_type, name);
+                obj_type = PyObject_Type(cfn->m_self);
+
+                // use method descriptor instead of instance methods for builtin
+                // objects. Othwerwise, there might be some errors since we INCREF
+                // on the bound method. See: https://github.com/sumerc/yappi/issues/60
+                method_descriptor = PyObject_GetAttr(obj_type, name);
+                if (method_descriptor) {
+                    pit->fn_descriptor = method_descriptor;
+                    Py_INCREF(method_descriptor);
+                }
+
+                // get name from type+name
+                mo = _PyType_Lookup((PyTypeObject *)obj_type, name);
                 Py_XINCREF(mo);
                 Py_XDECREF(obj_type);
                 Py_DECREF(name);
@@ -567,6 +613,12 @@ _ccode2pit(void *cco, uintptr_t current_tag)
             }
             PyErr_Clear();
         }
+
+        if (pit->fn_descriptor == NULL) {
+            pit->fn_descriptor = (PyObject *)cfn;
+            Py_INCREF(cfn);
+        }
+
         pit->name = PyStr_FromString(cfn->m_ml->ml_name);
         return pit;
     }
@@ -764,6 +816,16 @@ decr_rec_level(uintptr_t key)
     return 1;
 }
 
+
+static long long
+_ctx_tickcount(void) {
+    long long now;
+
+    now = tickcount();
+    current_ctx->last_seen = now;
+    return now;
+}
+
 static long long
 _get_frame_elapsed(void)
 {
@@ -791,12 +853,13 @@ _get_frame_elapsed(void)
         }
 
     } else {
-        result = tickcount() - ci->t0;
+        result = _ctx_tickcount() - ci->t0;
     }
 
     return result;
 }
 
+
 static int 
 _coro_enter(_pit *cp, PyFrameObject *frame)
 {
@@ -878,11 +941,14 @@ _call_enter(PyObject *self, PyFrameObject *frame, PyObject *arg, int ccall)
     _pit_children_info *pci;
     uintptr_t current_tag;
 
-    //printf("call ENTER:%s %s\n", PyStr_AS_CSTRING(frame->f_code->co_filename),
-    //                             PyStr_AS_CSTRING(frame->f_code->co_name));
-
+    // printf("call ENTER:%s %s\n", PyStr_AS_CSTRING(frame->f_code->co_filename),
+    //                              PyStr_AS_CSTRING(frame->f_code->co_name));
+    
     current_tag = _current_tag();
 
+    //printf("call ENTER:%s %s %d\n", PyStr_AS_CSTRING(frame->f_code->co_filename),
+    //                             PyStr_AS_CSTRING(frame->f_code->co_name), current_tag);
+
     if (ccall) {
         cp = _ccode2pit((PyCFunctionObject *)arg, current_tag);
     } else {
@@ -913,7 +979,7 @@ _call_enter(PyObject *self, PyFrameObject *frame, PyObject *arg, int ccall)
         return;
     }
 
-    ci->t0 = tickcount();
+    ci->t0 = _ctx_tickcount();
 
     incr_rec_level((uintptr_t)cp);
 
@@ -932,8 +998,8 @@ _call_leave(PyObject *self, PyFrameObject *frame, PyObject *arg, int ccall)
     int yielded = 0;
     pci = ppci = NULL;
 
-    //printf("call LEAVE:%s %s\n", PyStr_AS_CSTRING(frame->f_code->co_filename),
-    //                             PyStr_AS_CSTRING(frame->f_code->co_name));
+    // printf("call LEAVE:%s %s\n", PyStr_AS_CSTRING(frame->f_code->co_filename),
+    //                              PyStr_AS_CSTRING(frame->f_code->co_name));
 
     elapsed = _get_frame_elapsed();
 
@@ -943,13 +1009,10 @@ _call_leave(PyObject *self, PyFrameObject *frame, PyObject *arg, int ccall)
         return;
     }
 
-    //if (frame->f_code->co_flags & CO_GENERATOR) {
-        //printf("is a generator func.\n");
-    //}
-
-    // TODO: Comment
+    // if the function that the frame belongs is a coroutine, we check if we RETURN
+    // or YIELD from the coroutine to calculate the correct walltime
     if (IS_ASYNC(frame)) {
-        if (frame->f_stacktop) {
+        if (IS_SUSPENDED(frame)) {
             yielded = 1;
             if (get_timing_clock_type() == WALL_CLOCK) {
                 // In fact setting this zero means following:
@@ -1072,6 +1135,7 @@ _yapp_callback(PyObject *self, PyFrameObject *frame, int what,
                PyObject *arg)
 {
     PyObject *last_type, *last_value, *last_tb;
+    _ctx* tl_prev_ctx;
     PyErr_Fetch(&last_type, &last_value, &last_tb);
 
     //printf("call EVENT %d %s %s", what, PyStr_AS_CSTRING(frame->f_code->co_filename),
@@ -1084,9 +1148,22 @@ _yapp_callback(PyObject *self, PyFrameObject *frame, int what,
         goto finally;
     }
 
-    // do not profile if multi-threaded is off and the context is different than
+    if (ctx_type == GREENLET && get_timing_clock_type() == CPU_CLOCK) {
+        tl_prev_ctx = (_ctx*)(get_tls_key_value(tl_prev_ctx_key));
+
+        if (tl_prev_ctx != current_ctx) {
+            if (tl_prev_ctx) {
+                _pause_greenlet_ctx(tl_prev_ctx);
+                _resume_greenlet_ctx(current_ctx);
+            }
+            if (set_tls_key_value(tl_prev_ctx_key, current_ctx) != 0)
+                goto finally;
+        }
+    }
+
+    // do not profile if multi-context is off and the context is different than
     // the context that called start.
-    if (!flags.multithreaded && current_ctx != initial_ctx) {
+    if (!flags.multicontext && current_ctx != initial_ctx) {
         goto finally;
     }
 
@@ -1136,7 +1213,7 @@ finally:
     // there shall be no context switch happenning inside 
     // profile events and no concurent running events is possible
     if (current_ctx->ts_ptr != PyThreadState_GET()) {
-        //printf("call EVENT %d %s %s %p %p\n", what, PyStr_AS_CSTRING(frame->f_code->co_filename),
+        // printf("call EVENT %d %s %s %p %p\n", what, PyStr_AS_CSTRING(frame->f_code->co_filename),
         //                     PyStr_AS_CSTRING(frame->f_code->co_name), 
         //                     current_ctx->ts_ptr, PyThreadState_GET());
 
@@ -1147,6 +1224,45 @@ finally:
     return 0;
 }
 
+static void
+_pause_greenlet_ctx(_ctx *ctx)
+{
+    ydprintf("pausing context: %ld", ctx->id);
+    ctx->gl_state.paused = 1;
+    ctx->gl_state.paused_at = tickcount();
+}
+
+static void
+_resume_greenlet_ctx(_ctx *ctx)
+{
+    long long shift;
+    int i;
+
+    ydprintf("resuming context: %ld", ctx->id);
+
+    if (!ctx->gl_state.paused) {
+        return;
+    }
+
+    ctx->gl_state.paused = 0;
+    shift = tickcount() - ctx->gl_state.paused_at;
+    ctx->t0 += shift;
+
+    for (i = 0; i <= ctx->cs->head; i++) {
+        ctx->cs->_items[i].t0 += shift;
+    }
+
+    ydprintf("resuming context: %ld, shift: %lld", ctx->id, shift);
+}
+
+static _ctx *
+_bootstrap_thread(PyThreadState *ts)
+{
+    ts->use_tracing = 1;
+    ts->c_profilefunc = _yapp_callback;
+    return NULL;
+}
+
 static _ctx *
 _profile_thread(PyThreadState *ts)
 {
@@ -1160,7 +1276,7 @@ _profile_thread(PyThreadState *ts)
         ctx = _create_ctx();
         if (!ctx) {
             return NULL;
-        }    
+        }
         if (!hadd(contexts, ctx_id, (uintptr_t)ctx)) {
             _del_ctx(ctx);
             if (!flput(flctx, ctx)) {
@@ -1178,8 +1294,10 @@ _profile_thread(PyThreadState *ts)
     ctx->id = ctx_id;
     ctx->tid = ts->thread_id;
     ctx->ts_ptr = ts;
+    ctx->gl_state.paused = 0;
+    ctx->gl_state.paused_at = 0;
 
-    //printf("Thread profile STARTED. ctx=%p, ts_ptr=%p, ctx_id=%ld, ts param=%p\n", ctx, 
+    // printf("Thread profile STARTED. ctx=%p, ts_ptr=%p, ctx_id=%ld, ts param=%p\n", ctx, 
     //        ctx->ts_ptr, ctx->id, ts);
 
     return ctx;
@@ -1230,6 +1348,9 @@ _init_profiler(void)
         flctx = flcreate(sizeof(_ctx), FL_CTX_SIZE);
         if (!flctx)
             goto error;
+        tl_prev_ctx_key = create_tls_key();
+        if (!tl_prev_ctx_key)
+            goto error;
         yappinitialized = 1;
     }
     return 1;
@@ -1247,6 +1368,10 @@ error:
         fldestroy(flctx);
         flctx = NULL;
     }
+    if (tl_prev_ctx_key) {
+        delete_tls_key(tl_prev_ctx_key);
+        tl_prev_ctx_key = NULL;
+    }
 
     return 0;
 }
@@ -1263,9 +1388,7 @@ profile_event(PyObject *self, PyObject *args)
         return NULL;
     }
 
-    if (flags.multithreaded) {
-        _ensure_thread_profiled(PyThreadState_GET());
-    }
+    _ensure_thread_profiled(PyThreadState_GET());
 
     ev = PyStr_AS_CSTRING(event);
 
@@ -1313,8 +1436,8 @@ _start(void)
         return 0;
     }
 
-    if (flags.multithreaded) {
-        _enum_threads(&_profile_thread);
+    if (flags.multicontext) {
+        _enum_threads(&_bootstrap_thread);
     } else {
         _ensure_thread_profiled(PyThreadState_GET());
         initial_ctx = _thread2ctx(PyThreadState_GET());
@@ -1362,6 +1485,9 @@ clear_stats(PyObject *self, PyObject *args)
     fldestroy(flctx);
     flctx = NULL;
 
+    delete_tls_key(tl_prev_ctx_key);
+    tl_prev_ctx_key = NULL;
+
     yappinitialized = 0;
     yapphavestats = 0;
     ycurfuncindex = 0;
@@ -1414,7 +1540,7 @@ _ctxenumstat(_hitem *item, void *arg)
 
     efn = (PyObject *)arg;
 
-    cumdiff = _calc_cumdiff(tickcount(), ctx->t0);
+    cumdiff = _calc_cumdiff(ctx->last_seen, ctx->t0);
 
     exc = PyObject_CallFunction(efn, "((skkfk))", tcname, ctx->id, ctx->tid,
         cumdiff * tickfactor(), ctx->sched_cnt);
@@ -1428,7 +1554,7 @@ _ctxenumstat(_hitem *item, void *arg)
 }
 
 static PyObject*
-enum_thread_stats(PyObject *self, PyObject *args)
+enum_context_stats(PyObject *self, PyObject *args)
 {
     PyObject *enumfn;
 
@@ -1437,7 +1563,7 @@ enum_thread_stats(PyObject *self, PyObject *args)
     }
 
     if (!PyArg_ParseTuple(args, "O", &enumfn)) {
-        PyErr_SetString(YappiProfileError, "invalid param to enum_thread_stats");
+        PyErr_SetString(YappiProfileError, "invalid param to enum_context_stats");
         return NULL;
     }
 
@@ -1453,7 +1579,7 @@ enum_thread_stats(PyObject *self, PyObject *args)
 
 int _pit_filtered(_pit *pt, _ctxfuncenumarg *eargs)
 {
-    _func_stat_filter filter;
+    _fast_func_stat_filter filter;
 
     filter = eargs->enum_args->func_filter;
 
@@ -1480,7 +1606,6 @@ _pitenumstat(_hitem *item, void *arg)
     PyObject *children;
     _pit_children_info *pci;
     _ctxfuncenumarg *eargs;
-    uintptr_t tag;
 
     children = NULL;
     pt = (_pit *)item->val;
@@ -1489,7 +1614,7 @@ _pitenumstat(_hitem *item, void *arg)
     if (_pit_filtered(pt, eargs)) {
         return 0;
     }
-    
+
     // do not show builtin pits if specified
     if  ((!flags.builtins) && (pt->builtin)) {
         return 0;
@@ -1519,17 +1644,12 @@ _pitenumstat(_hitem *item, void *arg)
     if (pt->callcount == 0)
         pt->callcount = 1;
 
-    tag = 0;
-    if (eargs->enum_args->func_filter.tag) {
-        tag = PyLong_AsVoidPtr(eargs->enum_args->func_filter.tag);
-    }
-
     exc = PyObject_CallFunction(eargs->enum_args->enumfn, "((OOkkkIffIOkOkO))", 
                         pt->name, pt->modname, pt->lineno, pt->callcount,
                         pt->nonrecursive_callcount, pt->builtin, 
                         _normt(pt->ttotal), _normt(pt->tsubtotal),
                         pt->index, children, eargs->ctx->id, eargs->ctx->name,
-                        tag, pt->fn_descriptor);
+                        eargs->tag, pt->fn_descriptor);
 
     if (!exc) {
         PyErr_Print();
@@ -1549,11 +1669,12 @@ _tagenumstat(_hitem *item, void *arg)
     _htab *pits;
     uintptr_t current_tag;
     _ctxfuncenumarg *eargs;
-    _func_stat_filter filter;
+    _fast_func_stat_filter filter;
 
     current_tag = item->key;
     eargs = (_ctxfuncenumarg *)arg;
     filter = eargs->enum_args->func_filter;
+    eargs->tag = current_tag;
 
     if (filter.tag) {
         if (current_tag != PyLong_AsVoidPtr(filter.tag)) {
@@ -1575,6 +1696,7 @@ _ctxfuncenumstat(_hitem *item, void *arg)
 
     ext_args.ctx = (_ctx *)item->val; 
     ext_args.enum_args = (_ctxenumarg *)arg;
+    ext_args.tag = 0; // default
 
     filtered_ctx_id = ext_args.enum_args->func_filter.ctx_id;
     if (filtered_ctx_id) {
@@ -1594,7 +1716,7 @@ start(PyObject *self, PyObject *args)
     if (yapprunning)
         Py_RETURN_NONE;
 
-    if (!PyArg_ParseTuple(args, "ii", &flags.builtins, &flags.multithreaded))
+    if (!PyArg_ParseTuple(args, "ii", &flags.builtins, &flags.multicontext))
         return NULL;
 
     if (!_start())
@@ -1611,8 +1733,8 @@ stop(PyObject *self)
     Py_RETURN_NONE;
 }
 
-void 
-_filterdict_to_statfilter(PyObject *filter_dict, _func_stat_filter* filter)
+int
+_filterdict_to_statfilter(PyObject *filter_dict, _fast_func_stat_filter* filter)
 {
     // we use a _statfilter struct to hold the struct and not to always 
     // as the filter_dict to get its values for each pit enumerated. This
@@ -1621,8 +1743,25 @@ _filterdict_to_statfilter(PyObject *filter_dict, _func_stat_filter* filter)
 
     fv = PyDict_GetItemString(filter_dict, "tag");
     if (fv) {
+        PyLong_AsVoidPtr(fv);
+        if (PyErr_Occurred()) {
+            yerr("invalid tag passed to get_func_stats.");
+            filter->tag = NULL;
+            return 0;
+        }
         filter->tag = fv;
     }
+    fv = PyDict_GetItemString(filter_dict, "ctx_id");
+    if (fv) {
+        PyLong_AsVoidPtr(fv);
+        if (PyErr_Occurred()) {
+            yerr("invalid ctx_id passed to get_func_stats.");
+            filter->ctx_id = NULL;
+            return 0;
+        }
+        filter->ctx_id = fv;
+    }
+
     fv = PyDict_GetItemString(filter_dict, "name");
     if (fv) {
         filter->name = fv;
@@ -1631,10 +1770,8 @@ _filterdict_to_statfilter(PyObject *filter_dict, _func_stat_filter* filter)
     if (fv) {
         filter->modname = fv;
     }
-    fv = PyDict_GetItemString(filter_dict, "ctx_id");
-    if (fv) {
-        filter->ctx_id = fv;
-    }
+
+    return 1;
 }
 
 static PyObject*
@@ -1665,7 +1802,9 @@ enum_func_stats(PyObject *self, PyObject *args)
         return NULL;
     }
 
-    _filterdict_to_statfilter(filter_dict, &ext_args.func_filter);
+    if (!_filterdict_to_statfilter(filter_dict, &ext_args.func_filter)) {
+        return NULL;
+    }
 
     henum(contexts, _ctxfuncenumstat, &ext_args);
 
@@ -1752,6 +1891,35 @@ set_context_name_callback(PyObject *self, PyObject *args)
     Py_RETURN_NONE;
 }
 
+static PyObject *
+set_context_backend(PyObject *self, PyObject *args)
+{
+    _ctx_type_t input_type;
+    
+    if (!PyArg_ParseTuple(args, "i", &input_type)) {
+        return NULL;
+    }
+    
+    if (input_type == ctx_type)
+    {
+        Py_RETURN_NONE;
+    }
+    
+    if (yapphavestats) {
+        PyErr_SetString(YappiProfileError, "backend type cannot be changed while stats are available. clear stats first.");
+        return NULL;
+    }
+
+    if (input_type != NATIVE_THREAD && input_type != GREENLET)  {
+        PyErr_SetString(YappiProfileError, "Invalid backend type.");
+        return NULL;
+    }
+
+    ctx_type = input_type;
+
+    Py_RETURN_NONE;
+}
+
 static PyObject *
 set_test_timings(PyObject *self, PyObject *args)
 {
@@ -1797,6 +1965,16 @@ set_clock_type(PyObject *self, PyObject *args)
     Py_RETURN_NONE;
 }
 
+static PyObject *
+get_context_backend(PyObject *self, PyObject *args)
+{
+    if (ctx_type == GREENLET) {
+        return Py_BuildValue("s", "GREENLET");
+    }  else {
+        return Py_BuildValue("s", "NATIVE_THREAD");
+    }
+}
+
 static PyObject *
 get_clock_time(PyObject *self, PyObject *args)
 {
@@ -1865,20 +2043,20 @@ get_start_flags(PyObject *self, PyObject *args)
 {
     PyObject *result = NULL;
     PyObject *profile_builtins = NULL;
-    PyObject *profile_multithread = NULL;
+    PyObject *profile_multicontext = NULL;
     
     if (!yapphavestats) {
         Py_RETURN_NONE;
     }
 
     profile_builtins = Py_BuildValue("i", flags.builtins);
-    profile_multithread = Py_BuildValue("i", flags.multithreaded);
+    profile_multicontext = Py_BuildValue("i", flags.multicontext);
     result = PyDict_New();
     PyDict_SetItemString(result, "profile_builtins", profile_builtins);
-    PyDict_SetItemString(result, "profile_multithread", profile_multithread);
+    PyDict_SetItemString(result, "profile_multicontext", profile_multicontext);
     
     Py_XDECREF(profile_builtins);
-    Py_XDECREF(profile_multithread);
+    Py_XDECREF(profile_multicontext);
     return result;
 }
 
@@ -1911,7 +2089,8 @@ static PyMethodDef yappi_methods[] = {
     {"start", start, METH_VARARGS, NULL},
     {"stop", (PyCFunction)stop, METH_NOARGS, NULL},
     {"enum_func_stats", enum_func_stats, METH_VARARGS, NULL},
-    {"enum_thread_stats", enum_thread_stats, METH_VARARGS, NULL},
+    {"enum_context_stats", enum_context_stats, METH_VARARGS, NULL},
+    {"enum_thread_stats", enum_context_stats, METH_VARARGS, NULL},
     {"clear_stats", clear_stats, METH_VARARGS, NULL},
     {"is_running", is_running, METH_VARARGS, NULL},
     {"get_clock_type", get_clock_type, METH_VARARGS, NULL},
@@ -1922,6 +2101,8 @@ static PyMethodDef yappi_methods[] = {
     {"set_context_id_callback", set_context_id_callback, METH_VARARGS, NULL},
     {"set_tag_callback", set_tag_callback, METH_VARARGS, NULL},
     {"set_context_name_callback", set_context_name_callback, METH_VARARGS, NULL},
+    {"set_context_backend", set_context_backend, METH_VARARGS, NULL},
+    {"get_context_backend", get_context_backend, METH_VARARGS, NULL},
     {"_get_start_flags", get_start_flags, METH_VARARGS, NULL}, // for internal usage.
     {"_set_test_timings", set_test_timings, METH_VARARGS, NULL}, // for internal usage.
     {"_profile_event", profile_event, METH_VARARGS, NULL}, // for internal usage.
@@ -1973,9 +2154,11 @@ init_yappi(void)
     yapprunning = 0;
     paused = 0;
     flags.builtins = 0;
-    flags.multithreaded = 0;
+    flags.multicontext = 0;
     test_timings = NULL;
-    
+
+    SUPPRESS_WARNING(_DebugPrintObjects);
+
     if (!_init_profiler()) {
         PyErr_SetString(YappiProfileError, "profiler cannot be initialized.");
 #ifdef IS_PY3K
